from datetime import datetime
from enum import auto, Enum
from pathlib import Path
from typing import Optional, Any, Tuple

from gi.repository import Notify
from pyee import BaseEventEmitter


class NotificationEvent(Enum):
    CLOSED = auto()
    LEFT_BUTTON_CLICKED = auto()
    RIGHT_BUTTON_CLICKED = auto()


class NotificationButton(Enum):
    LEFT = 'left'
    RIGHT = 'right'


ButtonCaptionsType = Tuple[
    str,  # left
    str,  # right
]


class Notification:
    """
    Create notification with the specified urgency, message, icon.
    Add buttons and listenable events for notification closing / buttons clicking
    """
    __notification_button_to_event_map = {
        NotificationButton.LEFT.value: NotificationEvent.LEFT_BUTTON_CLICKED.value,
        NotificationButton.RIGHT.value: NotificationEvent.RIGHT_BUTTON_CLICKED.value,
    }

    def __init__(self,
                 title: str,
                 message: str,
                 icon: Path,
                 urgency: Optional[Notify.Urgency] = None,
                 listen_closed_event: bool = False) -> None:
        self.__raw_notification = Notify.Notification.new(title, message, str(icon))

        if urgency is not None:
            self.__raw_notification.set_urgency(urgency)

        self.last_shown: Optional[datetime] = None

        self.event = BaseEventEmitter()

        if listen_closed_event:
            self.__raw_notification.connect('closed', self.__on_closed)

    def __on_closed(self, notification: Notify.Notification) -> None:
        self.event.emit(NotificationEvent.CLOSED.value)

    def __on_button_clicked(self, notification: Notify.Notification, action_id: str, arg: Any) -> None:
        event = self.__notification_button_to_event_map[action_id]

        # TODO: send self to handler as an argument
        self.event.emit(event)

    def add_buttons(self, button_captions: ButtonCaptionsType) -> None:
        """Add two active buttons to notification - left and right"""
        left, right = button_captions

        self.__raw_notification.add_action(
            NotificationButton.LEFT.value,
            left,
            self.__on_button_clicked,
            None
        )

        self.__raw_notification.add_action(
            NotificationButton.RIGHT.value,
            right,
            self.__on_button_clicked,
            None
        )

    def show(self) -> None:
        self.__raw_notification.show()

        self.last_shown = datetime.now()



from pathlib import Path

from gi.repository import AppIndicator3 as AppIndicator
from gi.repository import Gtk

FALLBACK_ICON = 'face-monkey'


class TrayIcon:
    def __init__(self, app_id: str, icon: Path, menu: Gtk.Menu) -> None:
        self.menu = menu

        self.indicator = \
            AppIndicator.Indicator.new(app_id, FALLBACK_ICON, AppIndicator.IndicatorCategory.APPLICATION_STATUS)

        self.set_icon_if_exist(icon)

        self.indicator.set_status(AppIndicator.IndicatorStatus.ACTIVE)
        self.indicator.set_menu(self.menu)

    def set_icon_if_exist(self, icon: Path) -> None:
        if not icon.exists():
            return

        self.indicator.set_icon(str(icon))



class ApplicationInfo:
    """
    ApplicationInfo — is the one of the base concept of speaking-eye.
    List of ApplicationInfo is created during the reading speaking-eye config
    and later it is used for computing different stats that are displayed at reports.
    ApplicationInfo can be read with ApplicationInfoReader.

    ApplicationInfo is a more broad concept then Activity,
    because it operates with re expressions.
    They can be matched with each other using ApplicationInfoMatcher
    """

    def __init__(self, title: str, wm_name_re: str, tab_re: str, is_distracting: bool) -> None:
        self.title = title
        self.wm_name_re = wm_name_re
        self.tab_re = tab_re
        self.is_distracting = is_distracting

    def __eq__(self, other: object) -> bool:
        """
        Overrides the default implementation
        to use the object values instead of identifiers for comparison
        """
        if not isinstance(other, ApplicationInfo):
            return False

        if self.title != other.title:
            return False

        if self.wm_name_re != other.wm_name_re:
            return False

        if self.tab_re != other.tab_re:
            return False

        return self.is_distracting == other.is_distracting



from datetime import timedelta
from pathlib import Path
from random import randint
from typing import Union, cast

import i18n

from .language import Language


class Localizator:
    """Wrapper to work with i18n internalization package"""

    FormatValueType = Union[str, int, timedelta]

    def __init__(self, localization_dir: Path, language: Language) -> None:
        if not localization_dir.is_dir():
            raise ValueError(f'Path [{localization_dir}] does not exist or it is not a dir!')

        i18n.load_path.append(localization_dir)

        i18n.set('locale', language.value)
        i18n.set('enable_memoization', True)  # cache loaded strings in memory
        i18n.set('filename_format', '{locale}.{format}')

    def get(self, key: str, **kwargs: FormatValueType) -> str:
        return cast(str, i18n.t(key, **kwargs))

    def get_random(self, key: str, max_n: int, **kwargs: FormatValueType) -> str:
        variant = randint(1, max_n)
        key_with_variant = key + i18n.get('namespace_delimiter') + str(variant)

        return self.get(key_with_variant, **kwargs)



from datetime import date
from typing import List, Tuple

from .activity import Activity
from .activity_builder import ActivityBuilder
from .activity_helper import ActivityHelper
from .datetime_helper import DatetimeHelper
from .value import Value


class ActivitySplitter:
    """Split activity lasting several days into several activities"""

    @staticmethod
    def split_by_day(activity: Activity) -> List[Tuple[date, Activity]]:
        days = ActivityHelper.get_days(activity)
        first_day, last_day = days[0], days[-1]

        if first_day == last_day:
            return [(first_day, activity)]

        result = []

        for day in days:
            start_time, end_time = DatetimeHelper.get_date_range(day)

            new_activity = ActivityBuilder(activity)\
                .with_start_time(activity.start_time if day == first_day else start_time)\
                .with_end_time(
                    Value.get_or_raise(activity.end_time, 'activity.end_time') if day == last_day else end_time
                )\
                .build()

            result.append((day, new_activity))

        return result



from datetime import datetime, timedelta
from typing import Optional

from .application_info import ApplicationInfo


class Activity:
    """
    Activity — is the one of the base concept of speaking-eye.
    Any user action with a window will turn into Activity.
    In can be written to file with ActivityWriter or read from raw data with ActivityReader.
    Also activities are stored in ActivityStatHolder using ActivityStat to compute total time spent in all activities.

    Activity is a more specific concept then ApplicationInfo,
    so two or more activities with different wm_class or window_name
    can have the same ApplicationInfo
    """
    def __init__(self, wm_class: str, window_name: str, start_time: datetime, is_work_time: bool) -> None:
        self.wm_class = wm_class
        self.window_name = window_name
        self.start_time = start_time
        self.end_time: Optional[datetime] = None
        self.activity_time: Optional[timedelta] = None
        self.is_work_time = is_work_time
        self.application_info: Optional[ApplicationInfo] = None

    def set_end_time(self, end_time: datetime) -> 'Activity':
        if end_time < self.start_time:
            raise ValueError(f'end_time [{end_time}] should be greater than start_time [{self.start_time}]!')

        self.end_time = end_time
        self.activity_time = self.end_time - self.start_time

        return self

    def has_finished(self) -> bool:
        return self.end_time is not None

    def __eq__(self, other: object) -> bool:
        """
        Overrides the default implementation
        to use the object values instead of identifiers for comparison
        """
        if not isinstance(other, Activity):
            return False

        if self.wm_class != other.wm_class:
            return False

        if self.window_name != other.window_name:
            return False

        if self.start_time != other.start_time:
            return False

        if self.end_time != other.end_time:
            return False

        if self.activity_time != other.activity_time:  # pragma: no cover
            return False

        if self.is_work_time != other.is_work_time:
            return False

        return self.application_info == other.application_info

    def set_application_info(self, value: ApplicationInfo) -> 'Activity':
        self.application_info = value

        return self



import logging
from typing import Callable, Optional

from gi.repository import GLib, GObject


class Timer:
    def __init__(self, name: str,
                 handler: Callable[[], None],
                 interval_ms: int,
                 repeat: bool,
                 logger: logging.Logger) -> None:
        self.name = name
        self.handler = handler
        self.interval_ms = interval_ms
        self.repeat = repeat
        self.timer_id: Optional[int] = None
        self.is_started = False

        self.logger = logger

    def start(self) -> None:
        if self.is_started:
            self.logger.debug(f'Timer [{self.name}] has already started, thus will not start again')
            return

        self.timer_id = GLib.timeout_add(self.interval_ms, self.__internal_handler)
        self.is_started = True

    def stop(self) -> None:
        if not self.is_started:
            self.logger.debug(f'Timer [{self.name}] is not started, thus will not stop')
            return

        GObject.source_remove(self.timer_id)
        self.timer_id = None
        self.is_started = False

    def __internal_handler(self) -> bool:
        self.handler()

        if not self.repeat:
            self.is_started = False

        return self.repeat



from typing import Any, Dict, List, Mapping, Sequence, Type, TypeVar, Union, cast

from pydash import get
from typeguard import check_type

# declare type variable
T = TypeVar('T')


class TypedValue:
    @staticmethod
    def check_type(name: str, value: Any, value_type: Type[T]) -> T:
        check_type(name, value, value_type)

        return cast(T, value)

    # TODO: Optional or Union as value_type does not work
    @staticmethod
    def get(obj: Union[List[Any], Dict[Any, Any], Sequence[Any], Mapping[Any, Any]],
            path: str, value_type: Type[T], default_value: T) -> T:
        """
        Get a value from an object by path if the value has requested type.
        Raise exception otherwise
        """
        result = get(obj, path)

        if result is None:
            name = f'default_value [{default_value}] for path [{path}]'

            # TODO: remove default value runtime checking when mypy will handle such cases correctly
            #       For example, TypedValue.get(dict_with_abc, 'a.b.c', int, '6') should produce mypy type error:
            #       '6' (T) is not int (Type[T])

            return TypedValue.check_type(name, default_value, value_type)

        name = f'value [{result}] from path [{path}]'

        return TypedValue.check_type(name, result, value_type)



import glob
import os
from datetime import date, datetime
from pathlib import Path
from shutil import copy
from typing import cast, Optional, Union

import parse
from xdg import xdg_config_home, xdg_data_home

from .icon_state import IconState
from .theme import Theme
from .value import Value


class FilesProvider:
    """"
    Provide paths to localization / icon / data files.
    Help to search among files with raw data
    """
    __DATE_FORMAT_LABEL = 'date'
    __RAW_DATA_FILE_MASK = f'{{{__DATE_FORMAT_LABEL}}}_speaking_eye_raw_data.tsv'
    __FS_RAW_DATA_FILE_MASK = __RAW_DATA_FILE_MASK.format_map({__DATE_FORMAT_LABEL: '*'})

    def __init__(self, package_root_dir: Path, app_id: str) -> None:
        self.__package_root_dir = package_root_dir

        self.__i18n_dir = self.__package_root_dir / 'i18n'
        self.__icon_dir = self.__package_root_dir / 'icon'
        self.__initial_config_path = self.__package_root_dir / 'config' / 'config.yaml'

        self.__autostart_file_path = xdg_config_home() / 'autostart' / 'speaking-eye.desktop'
        data_home = xdg_data_home()
        self.__desktop_file_path = data_home / 'applications' / 'speaking-eye.desktop'

        app_data_dir = data_home / app_id
        self.__raw_data_dir = app_data_dir / 'data'
        self.__default_config_path = app_data_dir / 'config' / 'config.yaml'

        self.__raw_data_dir.mkdir(parents=True, exist_ok=True)

        self.__check_dirs()

        self.__config_path: Optional[Path] = None

    def __check_dirs(self) -> None:
        dir_paths = [
            self.__package_root_dir,
            self.__i18n_dir,
            self.__icon_dir,
            self.__raw_data_dir,
        ]

        for dir_path in dir_paths:
            if dir_path.is_dir():
                continue

            raise ValueError(f'Path [{dir_path}] does not exist or it is not a dir!')

    @property
    def i18n_dir(self) -> Path:
        return self.__i18n_dir

    @property
    def autostart_file_path(self) -> Path:
        return self.__autostart_file_path

    @property
    def desktop_file_path(self) -> Path:
        return self.__desktop_file_path

    @property
    def config_path(self) -> Path:
        return Value.get_or_raise(self.__config_path, '__config_path')

    @config_path.setter
    def config_path(self, raw_value: Union[Path, str]) -> None:
        value = raw_value

        if isinstance(raw_value, str):
            value = Path(raw_value)

        self.__config_path = cast(Path, value)

        if self.__config_path.exists():
            return

        is_custom_config_path = self.default_config_path != self.__config_path

        if is_custom_config_path:
            raise ValueError(f'__config_path [{self.__config_path}] does not exist!')

        self.__config_path.parent.mkdir(parents=True, exist_ok=True)
        copy(self.__initial_config_path, self.__config_path)

    @property
    def default_config_path(self) -> Path:
        return self.__default_config_path

    @property
    def raw_data_dir(self) -> Path:
        return self.__raw_data_dir

    def get_icon_file_path(self, theme: Theme, icon_state: IconState) -> Path:
        return self.__icon_dir / cast(str, theme.value) / f'{icon_state.value}.png'

    def get_raw_data_file_path(self, file_date: date) -> Path:
        return self.__raw_data_dir / self.__RAW_DATA_FILE_MASK.format_map({self.__DATE_FORMAT_LABEL: file_date})

    def get_date_of_first_raw_data_file(self, default_date: date = date.today()) -> date:
        """Return min date from raw data file names or default_date if no files"""
        file_paths = glob.glob(str(self.__raw_data_dir / self.__FS_RAW_DATA_FILE_MASK))

        if len(file_paths) == 0:
            return default_date

        file_names = [os.path.basename(file_path) for file_path in file_paths]

        dates = []
        for file_name in file_names:
            date_str = parse.parse(self.__RAW_DATA_FILE_MASK, file_name)[self.__DATE_FORMAT_LABEL]
            file_date = datetime.strptime(date_str, '%Y-%m-%d').date()

            dates.append(file_date)

        return min(dates)



from typing import Generic, TypeVar, Optional, Callable

# declare type variable
T = TypeVar('T')


class Value(Generic[T]):

    @staticmethod
    # NOTE: type T will be inherited based on the function call
    def get_or_raise(value: Optional[T], name: str) -> T:
        """
        Return value if it is not None.
        Raise exception otherwise
        """
        if value is None:
            raise ValueError(f'Value [{name}] is None!')

        return value

    @staticmethod
    def get_by_getter_or_default(value_getter: Callable[[], Optional[T]], default: T) -> T:
        """
        Return value from value_getter if all attributes exist and set.
        Return default value otherwise
        """
        try:
            value = value_getter()

            return value if value is not None else default

        except (AttributeError, NameError):
            return default



from datetime import timedelta
from typing import List, ItemsView, cast

from .activity import Activity
from .activity_helper import ActivityHelper
from .activity_stat import ActivityStat
from .application_info import ApplicationInfo
from .special_application_info_title import SpecialApplicationInfoTitle

KeyType = str
ItemType = ActivityStat
ActivityStatHolderItemsType = ItemsView[KeyType, ItemType]


class ActivityStatHolder(dict):  # type: ignore[type-arg]
    """Store ActivityStat objects and compute total time spent in all activities"""

    def __init__(self, activities: List[Activity]) -> None:
        super().__init__()

        self.total_work_time = timedelta()
        self.total_off_time = timedelta()

        for activity in activities:
            self.update_stat(activity)

    def initialize_stats(self, application_infos: List[ApplicationInfo]) -> None:
        """
        Add to holder all titles from app info list
        even if there was not activity in these apps to
        simplify further holder uses and stats analyses
        """
        for application_info in application_infos:
            if application_info.title in self:
                continue

            self[application_info.title] = ActivityStat()

    def update_stat(self, activity: Activity) -> None:
        if activity.is_work_time:
            self.total_work_time += ActivityHelper.get_activity_time(activity)
        else:
            self.total_off_time += ActivityHelper.get_activity_time(activity)

        title_from_config = activity.application_info.title if activity.application_info is not None \
            else SpecialApplicationInfoTitle.OTHERS.value

        if title_from_config not in self:
            self[title_from_config] = ActivityStat.from_activity(activity)

            return

        activity_stat = self[title_from_config]
        activity_stat.update(activity)

    def get_group_work_time(self, titles: List[str]) -> timedelta:
        """Compute total work time for the specific group of ActivityStat titles"""
        result = timedelta()

        for title in titles:
            if title not in self:
                raise RuntimeError(f'ActivityStatHolder does not contain stat for [ActivityStat.title={title}]!')

            result += self[title].work_time

        return result

    # NOTE: override the default implementations just to use our typing

    def __setitem__(self, app_name: KeyType, item: ItemType) -> None:
        super().__setitem__(app_name, item)

    def __getitem__(self, app_name: KeyType) -> ItemType:
        return cast(ItemType, super().__getitem__(app_name))

    def __contains__(self, app_name: KeyType) -> bool:  # type: ignore[override]
        return super().__contains__(app_name)

    def items(self) -> ActivityStatHolderItemsType:
        return super().items()



from datetime import timedelta, datetime, date
from typing import List

from .activity import Activity
from .activity_converter import ActivityConverter
from .datetime_helper import DatetimeHelper


class ActivityHelper:
    """Extra methods for working with Activity objects"""

    @staticmethod
    def raise_if_not_finished(activity: Activity) -> None:
        if not activity.has_finished():
            raise ValueError(f'Activity [{ActivityConverter.to_string(activity)}] should be finished!')

    @staticmethod
    def get_activity_time(activity: Activity) -> timedelta:
        ActivityHelper.raise_if_not_finished(activity)

        if activity.activity_time is None:
            raise ValueError(f'Activity [{ActivityConverter.to_string(activity)}] activity_time is None!')

        return activity.activity_time

    @staticmethod
    def get_end_time(activity: Activity) -> datetime:
        ActivityHelper.raise_if_not_finished(activity)

        if activity.end_time is None:
            raise ValueError(f'Activity [{ActivityConverter.to_string(activity)}] end_time is None!')

        return activity.end_time

    @staticmethod
    def get_days(activity: Activity) -> List[date]:
        """Get all dates between activity start day and its end day"""
        return DatetimeHelper.get_dates_between(activity.start_time.date(),
                                                ActivityHelper.get_end_time(activity).date())



import re
import subprocess


def get_wm_class(xid: int) -> str:
    """Get WM_CLASS window attribute from the xprop utility"""
    process_result = subprocess.run(['xprop', '-id', f'{xid}', 'WM_CLASS'], stdout=subprocess.PIPE)
    xprop_output = process_result.stdout.decode('utf-8')

    wm_class_re = r'WM_CLASS\(STRING\) = "(.+)", "(.+)"$'
    search = re.search(wm_class_re, xprop_output)

    if not search:
        raise Exception(f'WM_CLASS does not found in [{xprop_output}]')

    return search.group(2)



import os

from setuptools import setup

requires = (
    'coloredlogs>=14.0',
    'dash>=1.19.0',
    'pandas>=1.0.3',
    'parse>=1.19.0',
    'plotly>=4.14.3',
    'pydash>=4.8.0',
    'pyee>=7.0.2',
    'python-i18n>=0.3.9',
    'pyyaml>=5.4',
    'typeguard>=2.12.1',
    'xdg==5.1.1',
)

this_directory = os.path.abspath(os.path.dirname(__file__))


# NOTE: solution from https://packaging.python.org/guides/single-sourcing-package-version/
def get_version(rel_path: str) -> str:
    init_path = os.path.join(this_directory, rel_path)
    with open(init_path, 'r') as f:
        for line in f.read().splitlines():
            if line.startswith('__version__'):
                delim = '"' if '"' in line else "'"

                return line.split(delim)[1]

        raise RuntimeError(f'Unable to find version string in [{init_path}]!')


def get_long_description() -> str:
    descr_path = os.path.join(this_directory, 'doc/pypi/long_description.md')
    with open(descr_path, encoding='utf-8') as f:
        return f.read()


setup(
    name='speaking-eye',
    version=get_version('src/speaking_eye/__init__.py'),
    description="Track & analyze your computer activity. Focus on work and don't forget to take breaks",
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    url='https://github.com/alena-bartosh/speaking-eye',
    python_requires='>=3.6, <4',
    install_requires=requires,
    package_dir={'': 'src'},
    packages=['speaking_eye'],
    # NOTE: data_files will be copied into path with prefix ~/.local/ when installation without "sudo"
    #       It is expected behaviour
    data_files=[
        ('share/icons/hicolor/scalable/apps', ['src/speaking_eye/desktop/speaking-eye.png']),
        ('share/applications', ['src/speaking_eye/desktop/speaking-eye.desktop']),
    ],
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'speaking-eye = speaking_eye.__main__:main',
        ],
    },
    author='Alena Bartosh',
    author_email='alena.mathematics@gmail.com',
    keywords='worktracker timetracker gtk wnck dash',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
    ],
)



from typing import Dict, List

from .application_info import ApplicationInfo

WM_NAME_KEY = 'wm_name'
TAB_KEY = 'tab'

DataItemType = Dict[
    str,  # app_name
    Dict[
        str,  # wm_name or tab
        str  # value
    ]
]


class ApplicationInfoReader:
    """
    Try to read data items that consist of app name (title), wm_name, tab
    and create collection of ApplicationInfo
    """

    def __handle_regular_case(self, data_item: DataItemType, is_distracting: bool) -> ApplicationInfo:
        raw_app_info, = data_item.items()
        app_name, app_info = raw_app_info

        wm_name = app_info[WM_NAME_KEY] if WM_NAME_KEY in app_info else ''
        tab = app_info[TAB_KEY] if TAB_KEY in app_info else ''

        if not wm_name and not tab:
            raise ValueError(f'Application [{app_name}] has empty wm_name and tab!')

        return ApplicationInfo(app_name, wm_name, tab, is_distracting)

    def try_read(self, data: List[DataItemType], is_distracting: bool) -> List[ApplicationInfo]:
        if not isinstance(data, list):
            raise ValueError(f'Incorrect data type [{type(data)}]!')

        if len(data) == 0:
            raise ValueError('Empty data!')

        result = []

        for data_item in data:
            if not isinstance(data_item, dict):
                raise ValueError(f'Application info [{data_item}] in detailed/distracting list should be an object!')

            result.append(self.__handle_regular_case(data_item, is_distracting))

        return result



class BoolParser:
    """
    Parse string True/False to boolean.
    Raise exception otherwise
    """

    @staticmethod
    def parse(value: str) -> bool:
        if value == 'True':
            return True
        elif value == 'False':
            return False
        else:
            raise ValueError(f'Unexpected value [{value}]')
